/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  cache.c -- Stuff for requesting files through an external cache (eg. squid)

Overview:

   cache_init() redirects USER in ftp_cmds to come here. If user is
   anonymous we redirect all the other stuff we are going to need to
   parse or intercept for caching to come here too.

   We intercept RETR commands, and rewrite them in HTTP to retrieve
   through the proxy. We are called (cache_inc_data) with all
   server->client data transfers so we can strip HTTP headers etc.
   out.

  Current Problems include: 
      o No support for record structure.
      o STAT does nothing useful.
      o If we start doing anything with USER or 220 cmds then we have
        the potential to interact with non-transparent proxying code.
  ***************************************/


#include <sys/signal.h>
#include <sys/wait.h>
#include <fcntl.h>

#include "common.h"
#include "control.h"
#include "ftp-cmds.h"

void cache_user(const unsigned char *cmd,const unsigned char *arg);
void cache_stat(const unsigned char *cmd,const unsigned char *arg);
void cache_abor(const unsigned char *cmd,const unsigned char *arg);
void cache_mode(const unsigned char *cmd,const unsigned char *arg);
void cache_stru(const unsigned char *cmd,const unsigned char *arg);
void cache_type(const unsigned char *cmd,const unsigned char *arg);
void cache_rest(const unsigned char *cmd,const unsigned char *arg);
void cache_retr(const unsigned char *cmd,const unsigned char *arg);
void cache_stor(const unsigned char *cmd,const unsigned char *arg);

void got_size(int code, const unsigned char *msg);
void got_cwd(const unsigned char *msg);
unsigned char* read_httpline(void);

extern void closedata(int closeme);
extern void abor_parse(const unsigned char *cmd, const unsigned char * arg);

static RETSIGTYPE cache_signal(int signo);

static enum {
	NOTHING=FALSE,
	GOT_SIZE,
	GOT_CWD,
	CLOSE
} expecting = NOTHING;

static int user_ind, stat_ind, abor_ind, mode_ind;
static int stru_ind, type_ind, rest_ind, retr_ind, stor_ind;

static int working=FALSE;
static int httpstripped, httpsuccess;
static int type=0; /*Ascii or Binary*/
static int noop=0; /*No of NOOPs sent*/
static int range=0;
static unsigned char filename[MAX_LINE_LEN];

/* ------------------------------------------------------------- **
** Find and store the ftp_cmds that we may need to modify.
** Modify the ftp_cmds structure so that we get called on login.
** ------------------------------------------------------------- */
void cache_init(void)
{
	int i;

	for(i=0; *(ftp_cmds+i)->name!=0; i++){
		if(!strcasecmp((ftp_cmds+i)->name, "USER")) user_ind=i;
		else if(!strcasecmp((ftp_cmds+i)->name, "MODE")) mode_ind=i;
		else if(!strcasecmp((ftp_cmds+i)->name, "STRU")) stru_ind=i;
		else if(!strcasecmp((ftp_cmds+i)->name, "TYPE")) type_ind=i;
		else if(!strcasecmp((ftp_cmds+i)->name, "REST")) rest_ind=i;
		else if(!strcasecmp((ftp_cmds+i)->name, "RETR")) retr_ind=i;
		else if(!strcasecmp((ftp_cmds+i)->name, "STOR")) stor_ind=i;
		else if(!strcasecmp((ftp_cmds+i)->name, "STAT")) stat_ind=i;
		else if(!strcasecmp((ftp_cmds+i)->name, "ABOR")) abor_ind=i;
	}

	(ftp_cmds+user_ind)->cmd=cache_user;
}

/* ------------------------------------------------------------- **
** If login is anonymous then redirect the cmds we need to intercept
** ------------------------------------------------------------- */
void cache_user(const unsigned char *cmd, const unsigned char *arg)
{
	send_command(cmd, arg);

	if(strncasecmp(arg,"ftp",3) && strncasecmp(arg,"anonymous",9))
		return;

	(ftp_cmds+mode_ind)->cmd=cache_mode;
	(ftp_cmds+stru_ind)->cmd=cache_stru;
	(ftp_cmds+type_ind)->cmd=cache_type;
	(ftp_cmds+rest_ind)->cmd=cache_rest;
	(ftp_cmds+retr_ind)->cmd=cache_retr;
	(ftp_cmds+stor_ind)->cmd=cache_stor;
}

/* ------------------------------------------------------------- **
** Commands to intercept at all times
** ------------------------------------------------------------- */
void cache_mode(const unsigned char *cmd,const unsigned char *arg)
{
	if(!strncasecmp(arg, "S", 1)) send_message(200, "Command okay");
	else send_message(504,"Only stream mode implemented");
}

void cache_stru(const unsigned char *cmd,const unsigned char *arg)
{
	if(!strncasecmp(arg, "F", 1)) send_message(200, "Command okay");
	else send_message(504,"Only file structure implemented");
}

void cache_type(const unsigned char *cmd,const unsigned char *arg)
{
	if(!strncasecmp(arg, "I", 1)) {
		type=1;
		send_command(cmd,arg);
	}
	else if(!strncasecmp(arg, "AN", 1) || !strncasecmp(arg,"A",1)){
		type=0;
		send_command(cmd,arg);
	}
	else send_message(504,"Only types I and AN implemented");
}

void cache_rest(const unsigned char *cmd,const unsigned char *arg)
{
	/*We are in stream mode, so I think REST always gives a byte
	 * offset to start at. Could be wrong though*/
	debug("cache.c intercepted REST\n");
	range=atoi(arg);
	send_message(350,"Setting restart point. Send RETR to initiate transfer.");
}

void cache_stor(const unsigned char *cmd,const unsigned char *arg)
{
	if(range==0) send_command(cmd, arg);
	else send_message(504, "Can't do STOR after REST");
	range=0;
}

/* ------------------------------------------------------------- **
** Commands to intercept during transfer
** ------------------------------------------------------------- */
void cache_stat(const unsigned char *cmd,const unsigned char *arg)
{
	send_message(213, "Retrieving file through cache");
}

void cache_abor(const unsigned char *cmd,const unsigned char *arg)
{
	closedata(0);
	working=FALSE;

	debug("cache.c intercepted ABOR\n");
	send_message(426, "transfer aborted");
	send_message(226, "Closing data connection");
}

/* ------------------------------------------------------------- **
** And below here is the important bit. Intercept RETR commands,
** rewrite as HTTP, get from the HTTP proxy, and send back the file.
** ------------------------------------------------------------- */

/* ------------------------------------------------------------- **
** Determine the size of our file if we want.
** ------------------------------------------------------------- */
void cache_retr(const unsigned char *cmd,const unsigned char *arg)
{
	working=TRUE;
	strncpy(filename, arg, MAX_LINE_LEN-1);

	if(config_mincachesize()==0){
		got_size(213,"0");
		return;
	}

	send_command("SIZE", arg);
	expecting=GOT_SIZE;
}

/* ------------------------------------------------------------- **
** Send a PWD to get the working directory. We also need to set up
** commands for diverting during transfer.
** ------------------------------------------------------------- */
void got_size(int code, const unsigned char *msg)
{
	if(code!=213 || atoi(msg)<config_mincachesize()){
		working=FALSE;
		expecting=NOTHING;
		send_command("RETR", filename);
		debug("File too small for cache - retrieving directly\n");
		return;
	}
	debug("File size OK.\n");

	/*Get the working directory*/
	expecting=GOT_CWD;
	send_command("PWD","");

	(ftp_cmds+stat_ind)->cmd=cache_stat;
	(ftp_cmds+abor_ind)->cmd=cache_abor;
}

/* ------------------------------------------------------------- **
** And once we have the cwd get the file
** ------------------------------------------------------------- */
void got_cwd(const unsigned char *msg)
{
	unsigned char pwd[MAX_LINE_LEN], rangestr[MAX_LINE_LEN], *tp;
	unsigned char *buf=info->server_data.buf;
	char *host;
	int i;

	expecting=NOTHING;
        signal(SIGALRM, cache_signal);

	/* FIXME We ought to do some checking of the filename here
           before we pass it to squid*/

	/*Set up the HTTP command*/
	httpsuccess=httpstripped=FALSE;
	tp=strchr(msg,'"');
	if(tp==NULL) quit(SERVER_RUBBISH);

	strncpy(pwd,tp+1,MAX_LINE_LEN-1);

	tp=strchr(pwd,'"');
	if(tp==NULL) quit(SERVER_RUBBISH);
	*(tp++)='/';
	*tp=0;

	host = inet_ntoa(info->server_control.address.sin_addr);
	if(range>0) sprintf(rangestr, "Range: bytes=%d-\r\n", range);
	else rangestr[0]=0;
	i=snprintf(buf, BUF_LEN,
		   "GET ftp://%s/%s%s HTTP/1.0\r\n"
		   "Host: %s:%d\r\n"
		   "User-Agent: %s\r\n"
		   "%s\r\n",
		   host, pwd, filename, 
		   host, ntohs(info->server_control.address.sin_port),
		   "Frox/0.5",
		   rangestr );

	if(i<0 || i>BUF_LEN) quit(CLIENT_RUBBISH);

	debug2("HTTP string = %s\n", buf);

	if(info->state & SERVERDATA) close(info->server_data.fd);
	if( (info->server_data.fd=connect_to_socket(config_ftpproxy(), PASV))==-1){
		write_log(ERROR, "Unable to contact HTTP proxy");
		send_message(451, "Proxy error");
		working=FALSE;
		range=0;
		(ftp_cmds+stat_ind)->cmd=send_command;
		(ftp_cmds+abor_ind)->cmd=abor_parse;
		return;
	}
	send_message(150,"Starting transfer");

	/*And set everything up as if the connection was normal*/
	if( !(info->state & CLIENTDATA) ){
		if(info->state & ACTIVE){
			info->client_data.fd = 
				connect_to_socket(info->client_data.address, ACTIVE);
			info->state |= CLIENTDATA;
		}
	}

	info->client_data.buflen=0;
	info->server_data.buflen=strlen(buf);
	info->state |=SERVERDATA;

        fcntl(info->client_data.fd, F_SETFL, O_NONBLOCK);
        fcntl(info->server_data.fd, F_SETFL, O_NONBLOCK);	

	/*Finally set up our timer for keeping ftp server alive.*/
	alarm(30); /*FIXME make this configureable*/
}

/* ------------------------------------------------------------- **
** Called whenever we select on incoming data from server. Strip out
** HTTP lines, and if the request failed send a failure message and
** make sure we continue to ignore everything.
** ------------------------------------------------------------- */
void cache_inc_data(void)
{
	unsigned char *buf, *p;

	if(!working) return;
	if(httpstripped && !httpsuccess){ /*Discard message body*/
		debug("Discarding HTTP body after failure.\n");
		info->client_data.buflen=0;
		return;
	}
	if(httpstripped) return;
	
	/*Parse HTTP reply. FIXME - parse it properly!*/
	do{
		buf=read_httpline();
		if(buf==NULL) return; /*Incomplete line*/
		
		if( !strncmp(buf,"HTTP", 4)){ /*Status Line*/
			for(p=strchr(buf,' '); *p < '0' || *p > '9'; p++); 
			if(*p=='2' && (range==0 || *(p+2)=='6') ) httpsuccess=TRUE; /*Success*/
			else{ /*Failure - discard rest of message.*/
				info->client_data.buflen=0;
				httpsuccess=FALSE;
				httpstripped=TRUE;
				return;
			}
		}
		else if(strlen(buf)==0) httpstripped=TRUE;

	}while(!httpstripped);

	return;
}

void cache_close_data(void)
{
	alarm(0);
	if(!working && expecting!=CLOSE) return;
	working=FALSE;
	if(noop>0) {/*Wait until all our replies are in*/
		expecting=CLOSE;
		return; 
	}
	if(httpsuccess) send_message(226, "Transfer complete");
	else send_message(426, "Unable to transfer file");
	httpsuccess=httpstripped=FALSE;
	range=0;
	(ftp_cmds+stat_ind)->cmd=send_command;
	(ftp_cmds+abor_ind)->cmd=abor_parse;
	return;
}

/* ------------------------------------------------------------- **
** Deal with server replies if we need to. Note that we have to deal
** with replies to our NOOPs even if we aren't working as retrieval
** may finish in between sending NOOP and getting the reply.
** ------------------------------------------------------------- */
int cache_parsed_reply(int code, const unsigned char *msg)
{
	if(!config_usecache()) return(FALSE);
	if(noop>0){
		if(code>0) noop--;
		debug("Got NOOP reply\n");
		if(!working && noop==0) cache_close_data();
		return(TRUE);
	}
	if(expecting==NOTHING) return(FALSE);

	switch(expecting){
	case GOT_CWD:
		got_cwd(msg);
		break;
	case GOT_SIZE:
		got_size(code, msg);
		break;
	default:
	}
	return(TRUE);
}		

/* ------------------------------------------------------------- **
** Handle SIGALRMs to send ftp server keepalives.
** ------------------------------------------------------------- */
static RETSIGTYPE cache_signal(int signo)
{
	switch(signo){

	case SIGALRM: /*Send a NOOP to let the server know we are still here*/
		if(working) {
			alarm(30); /*FIXME make this configureable*/
			write(info->server_control.fd,"NOOP\r\n",6);
			debug("Sent NOOP\n");/*Ooops. printf in sighandler*/
			noop++;
		}
		break;
	}
}

/* ------------------------------------------------------------- **
** Remove one line from client_data buffer. return it (without
** terminating \r\n), or NULL if no complete line exists.
** ------------------------------------------------------------- */
unsigned char* read_httpline(void)
{
	static char buf[MAX_LINE_LEN];
	int i;

	for(i=0; i<info->client_data.buflen; i++) 
		if(info->client_data.buf[i]=='\r') break;
	if(info->client_data.buf[i]!='\r') return(NULL); /*Incomplete line*/	

	info->client_data.buf[i]=0;
	strncpy(buf, info->client_data.buf, MAX_LINE_LEN-1);
	buf[MAX_LINE_LEN-1]=0;

	i+=2;
	memmove(info->client_data.buf, info->client_data.buf+i, 
		info->client_data.buflen-i+1);
	info->client_data.buflen-=i;
	
	debug2("HTTP: %s\n",buf);
	return(buf);
}
