/***************************************
    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    common.h -- miscellaneous definitions.
					
***************************************/


#ifndef COMMON_H
#define COMMON_H    /*+ To stop multiple inclusions. +*/

#include "../config.h"

#include <stdio.h>
#include <sys/types.h>

#if STDC_HEADERS
# include <stdlib.h>
# include <string.h>
#elif HAVE_STRINGS_H
# include <strings.h>
#endif /* STDC_HEADERS */

#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif /* HAVE_SYS_TIME_H */
#endif /* TIME_WITH_SYS_TIME */

#include "systwrap.h"

/* Defines */
#ifdef DEBUG
#define debug(S) fprintf(stderr,S)
#define debug2(S,T) fprintf(stderr,S,T)
#else
#define debug(S)
#define debug2(S,T)
#endif

#define TRUE 1
#define FALSE 0

#define DM    242
#define IP    244
#define WILL  251
#define WONT  252
#define DO    253
#define DONT  254
#define IAC   255

#define BUF_LEN 4097
#define MAX_LINE_LEN 180  /* For control connections */

/* States - this is a bitfield. active/passive, whether we are listening 
on any ports, and whether we are connected to data ports on either client
or server.
 -------------------------------------------
| listen | clientdata | serverdata | active |
 -------------------------------------------
*/
#define ACTIVE            1
#define SERVERDATA        2
#define CLIENTDATA        4
#define CONNECTED         6
#define LISTEN            8

/* Error codes */
typedef enum _exitreason {
	TIMEOUT,          
	CLIENT_CLOSE,
	SERVER_CLOSE,      
	NO_CLIENT_DATA,       /* Unable to connect to it */
	NO_SERVER_DATA,    
	CLIENT_RUBBISH,       /* Malformed control stream */
	SERVER_RUBBISH,
	MALLOC,
	SOCKET
} exitreason;    

typedef enum _loglevel {
	INFO,                 /* Who connects, etc.*/
	ERROR,                /* Non critical errors*/
	ATTACK                /* Suspicious stuff */
} loglevel;

typedef enum _socketuse {
	ACTV,
	PASV,
	CTRL
} socketuse;

typedef struct _connection {
	struct sockaddr_in address;
	int fd;
	unsigned char buf[BUF_LEN];
	int buflen;
} connection;

typedef struct _session_info {
	connection client_control;
	connection server_control;
	connection client_data;
	connection server_data;

	int state;
	int listen; /*fd we are listening on*/
	int anon;
} session_info;

/*****************
**    misc.c    **
******************/
int listen_on_socket(struct sockaddr_in *listen_address, socketuse use);
int connect_to_socket(struct sockaddr_in listen_address, socketuse use);
struct sockaddr_in com2n(int a1,int a2,int a3,int a4,int p1,int p2);
void n2com(struct sockaddr_in address, int* a1,int* a2,int* a3,int* a4,
	   int* p1,int* p2);
void closeandexit(int status);
struct sockaddr_in extract_address(const unsigned char *buf);
void write_log(loglevel priority, unsigned char * msg, ...);
unsigned char *addr2name(struct in_addr address);
void quit(exitreason reason);
void do_linecontrol(unsigned char *buf);

/*****************
**    linux.c   **
******************/
int get_orig_dest(int fd, struct sockaddr *addr);
int get_local_address(struct sockaddr_in *addr);
int transp_connect(struct sockaddr_in dest, struct sockaddr_in src);
int tc_close(int fd);
int intercept_listen(struct sockaddr_in intercept,struct sockaddr_in listen_on, socketuse use);
int il_close(int fd);

/*****************
**   config.c   **
******************/
int process_cmdline(int argc, char *argv[]);
int read_config(void);
int reread_config(int *listening);
int config_bindtodevice(int fd);
int config_chroot(void);
int config_droppriv(void);
int config_openlog(void);
struct sockaddr_in config_getaddress(void);
int config_loport(socketuse use);
int config_hiport(socketuse use);
int config_maxforks(void);
int config_usesameaddress(void);
const char * config_pidfile(void);
int config_bouncedefend(void);
int config_ntp(void);
int config_apconv(void);
int config_usecache(void);
struct sockaddr_in config_ftpproxy(void);
int config_mincachesize(void);
struct timeval config_timeout(void);
/*TODO*/
#define config_transdata() FALSE

/*****************
**     ntp.c    **
******************/
#ifdef NON_TRANS_PROXY
 void ntp_changedest(void);
 int ntp_senduser(int code);
#else
 static inline void ntp_changedest(void){};
 static inline int ntp_senduser(int code){return FALSE;};
#endif

/*****************
**    cache.c   **
******************/
#ifdef USE_CACHE
 void cache_init(void);
 int cache_parsed_reply(int code, const unsigned char *msg);
 void cache_inc_data(void);
void cache_close_data(void);
#else
 static inline void cache_init(void){};
 static inline int cache_parsed_reply(int code, const unsigned char *msg){return FALSE;};
 static inline void cache_inc_data(void){};
static inline void cache_close_data(void){};
#endif

/*******************
**Global Variables**
********************/ 
session_info *info;

#endif /* COMMON_H */
