/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  config.c: configuration file parsing
  
  ***************************************/


#include <sys/stat.h>
#include <fcntl.h>
#include <netdb.h>
#include <pwd.h>
#include <grp.h>
#include <syslog.h>

#include "common.h"

void usage(int exitval);

struct options {
	unsigned char config_file[251];
	unsigned char listen_name[251];
	struct sockaddr_in listen_address;
	unsigned char device[251];
	unsigned char user[251];
	unsigned char group[251];
	uid_t uid;
	gid_t gid;
	unsigned char chroot[251];
	u_int16_t hippasv,loppasv; /* Port ranges to use */
	u_int16_t hipactv,lopactv;
	u_int16_t hipcont,lopcont;
	int maxforks;
	struct timeval timeout;
	int same_address;
	int ntp;
	int apconv;
	int bounce_defend;
	unsigned char logfile[251];
	char pidfile[251];
	struct sockaddr_in ftp_proxy;
	int mincache;
};

int parse_file(struct options * conf);

static struct options config;
static int chrooted=FALSE;

/* ------------------------------------------------------------- **
** Functions to read / parse the config file.
** ------------------------------------------------------------- */
int process_cmdline(int argc, char *argv[])
{
	int opt;

	/*Set up defaults*/
	memset(&config, 0, sizeof(config));
	strcpy(config.config_file,CONFIG_FILE);
	config.loppasv=config.lopactv=config.lopcont=40000;
	config.hippasv=config.hipactv=config.hipcont=50000;
	config.timeout.tv_sec=3600;
	config.listen_address.sin_addr.s_addr = htonl(INADDR_ANY);
	config.listen_address.sin_family =  AF_INET;
	config.ftp_proxy.sin_family = AF_INET;

	while((opt=getopt(argc,argv, "hf:")) !=-1) {
		switch(opt) {
		case 'f':
			strncpy(config.config_file, optarg, 250);
			break;
		case 'h':
			usage(0);
		case ':':
			fprintf(stderr, "Option -%c needs an argument\n", optopt);
			usage(-1);
		case '?':
			fprintf(stderr, "Unknown option -%c\n", optopt);
			usage(-1);
		}
	}
	if (optind!=argc) return(-1);
	return(0);
}

void usage(int exitval)
{	
	fprintf(stderr,"Usage frox [-f config_file] [-h]\n");

	exit(exitval);
}


int read_config(void)
{
       return(parse_file(&config));
}

int parse_file(struct options * conf)
{
	FILE *fp;
	unsigned char buf[BUF_LEN];
	unsigned char *p;

	
	if ( (fp=fopen(conf->config_file,"r")) == NULL ) return(-1);

	do{
		/* 
		** Remove leading whitespace. Skip comments and empty lines.
		*/
		if(fgets(buf,BUF_LEN,fp)==NULL) break;
		for(p=buf;*p==' ' || *p=='\t';p++);
		if(*p=='#') continue;
		memmove(buf, p, strlen(p) +1);
		if ((p = strchr(buf, '\n')) != NULL) *p = '\0';
		if(*buf=='\0') continue;

		p=strtok(buf, " \t=");

		if(!strncasecmp(p, "Listen", 20))
			strncpy(conf->listen_name, strtok(NULL, " \t="), 250);

		else if(!strncasecmp(p, "Port", 20))
			conf->listen_address.sin_port
				= htons(strtol(strtok(NULL, " \t="), NULL,10));
		
                else if(!strncasecmp(p, "BindToDevice", 20))
			strncpy(conf->device, strtok(NULL, " \t="), 250);

		else if(!strncasecmp(p, "User", 20))
			strncpy(conf->user, strtok(NULL, " \t="), 250);

		else if(!strncasecmp(p, "Group", 20))
			strncpy(conf->group, strtok(NULL, " \t="), 250);

                else if(!strncasecmp(p, "Chroot", 20))
                        strncpy(conf->chroot, strtok(NULL, " \t="), 250);

                else if(!strncasecmp(p, "LogFile", 20))
                        strncpy(conf->logfile, strtok(NULL, " \t="), 250);

		else if(!strncasecmp(p, "PidFile", 20))
			strncpy(conf->pidfile, strtok(NULL, " \t="), 250);

                else if(!strncasecmp(p, "MaxForks", 20))
			conf->maxforks=atoi(strtok(NULL, " \t="));

                else if(!strncasecmp(p, "Timeout", 20)){
			conf->timeout.tv_sec=atoi(strtok(NULL, " \t="));
			conf->timeout.tv_usec=0;
		}
                else if(!strncasecmp(p, "PassivePorts", 20)){
			conf->loppasv=atoi(strtok(NULL, " \t-"));
			conf->hippasv=atoi(strtok(NULL, " \t"));
                }
                else if(!strncasecmp(p, "ActivePorts", 20)){
			conf->lopactv=atoi(strtok(NULL, " \t-"));
			conf->hipactv=atoi(strtok(NULL, " \t"));
		}
                else if(!strncasecmp(p, "ControlPorts", 20)){
			conf->lopcont=atoi(strtok(NULL, " \t-"));
			conf->hipcont=atoi(strtok(NULL, " \t"));
		}
                else if(!strncasecmp(p, "SameAddress", 20))
		{
			if(!strncasecmp(strtok(NULL, " \t="), "yes", 4) ) 
				conf->same_address=TRUE;
			else conf->same_address=FALSE;
		}
                else if(!strncasecmp(p, "DoNTP", 20))
		{
			if(!strncasecmp(strtok(NULL, " \t="), "yes", 4) ) 
				conf->ntp=TRUE;
			else conf->ntp=FALSE;
		}
                else if(!strncasecmp(p, "APConv", 20))
		{
			if(!strncasecmp(strtok(NULL, " \t="), "yes", 4) ) 
				conf->apconv=TRUE;
			else conf->apconv=FALSE;
		}
                else if(!strncasecmp(p, "BounceDefend", 20))
		{
			if(!strncasecmp(strtok(NULL, " \t="), "yes", 4) ) 
				conf->bounce_defend=TRUE;
			else conf->bounce_defend=FALSE;
		}
                else if(!strncasecmp(p, "FtpProxy", 20)){
			inet_aton(strtok(NULL, ":"), &conf->ftp_proxy.sin_addr);
			conf->ftp_proxy.sin_port
				= htons(strtol(strtok(NULL, " \t="), NULL,10));
		}
                else if(!strncasecmp(p, "MinCacheSize", 20))
			conf->mincache=atoi(strtok(NULL, " \t="));


		else {
			fprintf(stderr, "Unrecognised config command %s\n",p);
			return(-1);
		}

	}while(!feof(fp));

	fclose(fp);

	if(*conf->listen_name == 0) conf->listen_address.sin_addr.s_addr = INADDR_ANY;
	else if(*conf->listen_name < '0' || *conf->listen_name >'9' ) {
		struct hostent *hostinfo;

		hostinfo = gethostbyname(conf->listen_name);
		if(hostinfo == NULL){
			fprintf(stderr, "Unable to resolve host %s\n", conf->listen_name);
			return(-2);
		}
		conf->listen_address.sin_addr.s_addr 
			= ((struct in_addr *) *hostinfo->h_addr_list)->s_addr;
	}
	else inet_aton(conf->listen_name, &conf->listen_address.sin_addr);

       if(config.chroot[strlen(config.chroot)-1] == '/')
               config.chroot[strlen(config.chroot)-1] = 0;

	/* We need to find out the uid / gid now. We must chroot
	** before changing priveleges, but if we wait until after
	** chroot we cannot parse /etc/passwd. 
	*/

	if(*conf->group!=0){
		struct group* grp;
		if( (grp=getgrnam(conf->group)) ==NULL){
			fprintf(stderr, "Unable to find group %s\n",conf->group);
			return(-3);
		}
		conf->gid=grp->gr_gid;
	}

	if(*conf->user!=0){
		struct passwd* pw;
		if( (pw=getpwnam(conf->user)) == NULL){
			fprintf(stderr, "Unable to find user %s\n",conf->user);
			return(-2);
		}
		conf->uid=pw->pw_uid;
	}
	return(0);
}	

/* ------------------------------------------------------------- **
** Reread the configfile. A bit messy - maybe we should limit the
** things that can be changed by changing config file and sending HUP?
** ------------------------------------------------------------- */
int reread_config(int *listening)
{
	int i;
	struct options newconfig;
	
	/*Set up defaults*/
	write_log(INFO, "Recieved SIGHUP - rereading config file");
	memset(&newconfig, 0, sizeof(newconfig));
	strcpy(newconfig.config_file,config.config_file);
	newconfig.loppasv=newconfig.lopactv=newconfig.lopcont=40000;
	newconfig.hippasv=newconfig.hipactv=newconfig.hipcont=50000;
	
	newconfig.timeout.tv_sec=3600;
	newconfig.listen_address.sin_family =  AF_INET;
	newconfig.listen_address.sin_addr.s_addr = htonl(INADDR_ANY);
	
	i=parse_file(&newconfig);
	if(i==-1) return(-1);
	if(i==-3 && !chrooted) write_log(ERROR, "Unable to determine UID/GID to drop to.");
	if(chrooted) strcpy(newconfig.chroot, config.chroot);
	
	i = (config.listen_address.sin_port==newconfig.listen_address.sin_port &&
	     config.listen_address.sin_addr.s_addr==newconfig.listen_address.sin_addr.s_addr &&
	     !strcmp(config.device, newconfig.device));
	config=newconfig;
	
	if(!chrooted) config_chroot();  /*No point re chrooting*/
	if(!i){ /*Let's not close any sockets we don't have to*/
		int newfd;
		
		newfd=listen_on_socket(&config.listen_address, CTRL);
		if(newfd!=-1){
			close(*listening);
			*listening=newfd;
		} else {
			write_log(ERROR,"Ooops - can't change socket.");
		}
		config_bindtodevice(*listening);
	}
	config_openlog();
	config_droppriv();
	
	write_log(INFO, "Config file successfully reread");
	return(0);
}

/* ------------------------------------------------------------- **
** Functions which implement config file settings.
** ------------------------------------------------------------- */
int config_bindtodevice(int fd)
{
	if(*config.device==0) return(0);
	if(setsockopt(fd, SOL_SOCKET, SO_BINDTODEVICE, 
		      (void *)config.device, (socklen_t) strlen(config.device)+1) != 0){
		perror("Binding to device");
		return(-1);
	}
	write_log(INFO, "Bound to device %s",config.device);
	return(0);
}
	
int config_chroot()
{
	if(*config.chroot==0) return(0);
	if( chroot(config.chroot)!=0 ) return(-1);
	if( chdir("/") !=0) return(-1);
	write_log(INFO, "Chrooted to %s", config.chroot);
	chrooted=TRUE;
	
	/*Rewrite our filenames so they still work*/
	if(strncmp(config.chroot, config.pidfile, strlen(config.chroot)))
		*config.pidfile=0;
	else memmove(config.pidfile, config.pidfile+strlen(config.chroot),
		     strlen(config.pidfile) - strlen(config.chroot) + 1);
	
	if(strncmp(config.chroot, config.logfile, strlen(config.chroot)))
		*config.logfile=0;
	else memmove(config.logfile, config.logfile+strlen(config.chroot),
		     strlen(config.logfile) - strlen(config.chroot) + 1);

	if(strncmp(config.config_file, config.logfile, strlen(config.config_file)))
		*config.logfile=0;
	else memmove(config.logfile, config.logfile+strlen(config.config_file),
		     strlen(config.logfile) - strlen(config.config_file) + 1);
	
	debug("Chrooted:\n");
	debug2("  Now pidfile is %s\n", config.pidfile);
	debug2("  Now logfile is %s\n", config.logfile);
	return(0);
}

int config_droppriv()
{
	if(config.gid!=0){
		setgid(config.gid);
		setgid(config.gid);
	}
	if(config.uid!=0){
		setuid(config.uid);
		setuid(config.uid);
		write_log(INFO, "Dropped priveledges");
	}
	return(0);
}

int config_openlog()
{
	if(*config.logfile!=0){
		int tmpfd;
		tmpfd=open(config.logfile, O_APPEND|O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR);
		if(tmpfd==-1 || dup2(tmpfd,2)==-1){
			fprintf(stderr, "Unable to open logfile %s\n",config.logfile);
			return(-1);
		}
	}
	openlog("frox", LOG_PID|LOG_CONS|LOG_NDELAY, LOG_DAEMON);
	return(0);
}

/* ------------------------------------------------------------- **
** Functions to return config file options.
** ------------------------------------------------------------- */
struct sockaddr_in config_getaddress()
{
	return(config.listen_address);
}

int config_loport(socketuse use)
{
	int ret=40000;
	
	switch(use) {
	case PASV:
		ret=config.loppasv;
		break;
	case ACTV:
		ret=config.lopactv;
		break;
	case CTRL:
		ret=config.lopcont;
		break;
	}
	
	return(ret);
}

int config_hiport(socketuse use)
{
	int ret=50000;
 
	switch(use) {
	case PASV:
		ret=config.hippasv;
		break;
	case ACTV:
		ret=config.hipactv;
		break;
	case CTRL:
		ret=config.hipcont;
		break;
	}

	return(ret);
}	
	
int config_maxforks()
{
	return(config.maxforks);
}

const char * config_pidfile()
{
	       return(config.pidfile);
}

struct timeval config_timeout(void)
{
	return(config.timeout);
}

int config_apconv(void)
{
	return(config.apconv);
}

int config_ntp(void)
{
	return(config.ntp);
}

int config_usesameaddress()
{
	return(config.same_address);
}

int config_bouncedefend()
{
	return(config.bounce_defend);
}

int config_usecache()
{
	if(config.ftp_proxy.sin_port==0) return(FALSE);
	else return(TRUE);
}

struct sockaddr_in config_ftpproxy()
{
	return(config.ftp_proxy);
}

int config_mincachesize()
{
	return(config.mincache);
}
