/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  control.c -- forward and parse the control stream.
  
  ***************************************/


#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

#include "control.h"
#include "common.h"
#include "ftp-cmds.h"

extern session_info *info;

int extract_clientcmd(unsigned char * buf, unsigned char * cmd, unsigned char * arg);
int extract_servercmd(unsigned char * buf, int * code, unsigned char * msg);

void parse_client_cmd(unsigned char *cmd, unsigned char * arg);
void parse_server_cmd(int code, unsigned char * msg);

void write_all(int fd, unsigned char *buf);

/* ------------------------------------------------------------- **
** Forward control stream data from client -> server. Non blocking 
** ------------------------------------------------------------- */
void client_control_forward()
{
	int i, j;
	unsigned char cmdbuf[5], argbuf[MAX_LINE_LEN];

	do{
		ioctl(info->client_control.fd, FIONREAD, &i);
	}while(i<0 && errno==EINTR);

	if(i<1) quit(CLIENT_CLOSE);

	j=strlen(info->client_control.buf);
	i = iread(info->client_control.fd, info->client_control.buf+j,
		 (i > (BUF_LEN-1) - j ? (BUF_LEN-1) - j : i) );
	info->client_control.buf[i+j]=0;

	/*buf is NULL terminated, but could contain anything*/
					 
	while( (i=extract_clientcmd(info->client_control.buf, cmdbuf, argbuf)) == 0 )
		parse_client_cmd(cmdbuf, argbuf);

	if(i<0) {
		send_message(500,"You're sending me rubbish. Goodbye!");
		quit(CLIENT_RUBBISH);
	}
}

/* ------------------------------------------------------------- **
** Forward control stream data from server -> client. Non blocking 
** ------------------------------------------------------------- */
void server_control_forward()
{
	int i, j, code;
	unsigned char argbuf[MAX_LINE_LEN];
	
	do{
		ioctl(info->server_control.fd, FIONREAD, &i);
	}while(i<0 && errno==EINTR);

	if(i<1) quit(SERVER_CLOSE);

	j=strlen(info->server_control.buf);

	i = iread(info->server_control.fd, info->server_control.buf+j, 
		 (i > (BUF_LEN-1)-j ? (BUF_LEN-1)-j : i) );
	info->server_control.buf[i+j]=0;

	/*buf is NULL terminated, but could contain anything*/

        while( (i=extract_servercmd(info->server_control.buf, &code, argbuf)) == 0 )
		parse_server_cmd(code,argbuf);

        if(i<0) {
		send_message(500,"FTP server is sending you rubbish! Closing connection.");
		quit(SERVER_RUBBISH);
	}
}

/***************************************************************************
 *
 * Functions which read the raw control stream --- be careful
 *
 **************************************************************************/

/* ------------------------------------------------------------- **
** Removes one line of control stream from buf (up to '\r\n'). <= 5
** chars are returned in cmd (up to 4 + \0), and <= MAX_LINE_LEN in
** arg. Must accept any NULL terminated buf and give sane output.
**  
** returns 0 on success, 1 if there is not a complete line in
** buf, and -X on non-sane buf. contents of buf unchanged on
** return(1), undefined on return(-X)
** ------------------------------------------------------------- */
int extract_clientcmd(unsigned char *buf, unsigned char *cmd, unsigned char *arg)
{
	unsigned char *p,*cmdend,*argstart,*end;

	cmdend=argstart=end=p=strpbrk(buf,"\r\n \t");
	if(p==NULL) return(1);

	if(*p==' ' || *p=='\t'){
		for(argstart=p;*argstart==' ' || *argstart=='\t';argstart++);
		end=p=strpbrk(p, "\r\n");
	}
	if(p==NULL) return(1);
	if(*p=='\r')
		if( (end=strpbrk(p,"\n"))==NULL ) return(1);
	
	/* We have a complete command. Start copying it over.
	   eg. "USER   ftp\r\r\r\n"  (Microsoft servers don't do \r\n !)
	   buf__|   |  |  |     |__end
        cmdend______|  |  |______p
	argstart_______|               */

	*cmdend=0;
	*p=0;

	if(buf[0]==IAC) do_linecontrol(buf);  /* Telnet cmd */

	if(strlen(buf)>4 || strlen(argstart) > MAX_LINE_LEN-1 ) {
		debug2("Client sent us \"%s\", ",buf);
		debug2("\"%s\"\n",argstart);
		return(-1);
	}
	strcpy(cmd, buf);
	strcpy(arg,argstart);
	
	memmove(buf, end+1, strlen(end+1)+1);
	return(0);
}

/* ------------------------------------------------------------- **
** As extract_clientcmd, but returns the code as an int.
** If code==0 then we are in a multiline. code<0 means we are
** starting a multiline. 
** ------------------------------------------------------------- */
int extract_servercmd(unsigned char *buf, int * code, unsigned char *msg)
{
	static int multiline=0;
	int i;
	unsigned char *p,*end;

	end=p=strpbrk(buf,"\r\n");

	if(p==NULL) return(1);
	if(*p=='\r')
		if( (end=strpbrk(p,"\n"))==NULL ) return(1);
	
	/* We have a complete command. Start copying it over.
	   eg. "501  go away now\r\r\n"  (Microsoft servers don't do \r\n !)
	   buf__|               |   |__end
	                        |______p       */

	*p=0;

	if(multiline == 0){
		for(i=0;i<3;i++) if( buf[i] < '0' || buf[i] > '9') return(-1);
		if(buf[3]!=' ' && buf[3]!='-') return(-1);
		*code=atoi(buf);
		if(buf[3] == '-') {  /* Start of a multiline */

			multiline = *code;
			*code=-*code;
		}
		if(strlen(buf+4) > MAX_LINE_LEN-1 ) return(-1);
		strcpy(msg,buf+4);
	} else {
		if(isdigit(buf[0]) && isdigit(buf[1]) && isdigit(buf[2])
		   && buf[3]==' ' && (*code=atoi(buf)) == multiline ) { /* End multiline */
			multiline=0;
			if(strlen(buf+4) > MAX_LINE_LEN-1) return(-1);
			strcpy(msg,buf+4);
		} else {             /* Continue Multiline */
			*code=0;
			if(strlen(buf) > MAX_LINE_LEN-1) return(-1);
			strcpy(msg,buf);
		}
	}

	memmove(buf, end+1, strlen(end+1)+1);
	return(0);
}

/* ------------------------------------------------------------- **
** Decide what to do with the command.
** ------------------------------------------------------------- */
void parse_client_cmd(unsigned char *cmd,unsigned char *arg)
{
	cmd_struct *p;

#if (CHECK_LEVEL == 1) /* Check for dodgy characters */
        int i;

	for(i=0;i<strlen(arg);i++)
		if(arg[i]>126 || arg[i]<32) arg[i]='?';
#endif

	for(p=ftp_cmds; *p->name!=0; p++){
		if(!strcasecmp(p->name, cmd)){
			p->cmd(cmd, arg);
			return;
		}
	}
	debug2("Command %s not implemented\n", cmd);
	send_message(502, "Command not implemented.");
}

void parse_server_cmd(int code, unsigned char * msg)
{
	int i;

	for(i=0;i<strlen(msg);i++)
		if(msg[i]>126 || msg[i]<32) msg[i]='?';

	if (!cache_parsed_reply(code,msg)) {
		switch(code)
		{
		case 227:
			pasv_reply(msg);
			break;
		case 220:
		case -220:
		case 0:     /*Multiline greetings confuse ntp stuff otherwise*/
			if(!ntp_senduser(code)) send_message(code, msg);
			break;
		default:
			send_message(code, msg);
		}
	}
	
}

/* ------------------------------------------------------------- **
** Send command to server 
** ------------------------------------------------------------- */
void send_command(const unsigned char * cmd, const unsigned char *arg )
{
	unsigned char buf[BUF_LEN];

	if(strlen(arg)==0)
		snprintf(buf,BUF_LEN-1, "%s\r\n",cmd);
	else    snprintf(buf,BUF_LEN-1,"%s %s\r\n",cmd, arg);

	write_all(info->server_control.fd, buf);

	debug2("  C: \033[31m%s\033[37m",buf); /*Ooooh Colour*/
}
		
/* ------------------------------------------------------------- **
** Write message to client. Don't give unchecked data!
** ------------------------------------------------------------- */
void send_message(int code, const unsigned char *msg)
{
	unsigned char buf[BUF_LEN];

	if(code>0) snprintf(buf,BUF_LEN-1,"%d %s\r\n",code, msg);
	else if(code<0) snprintf(buf,BUF_LEN-1,"%d-%s\r\n",-code, msg); 
	else snprintf(buf, BUF_LEN-1, "%s\r\n", msg);

	write_all(info->client_control.fd, buf);

	debug2("  S: \033[34m%s\033[37m",buf);	
}

/* ------------------------------------------------------------- **
** Write buf->fd. If fd has O_NONBLOCK set we will keep trying
** until TIMEOUT.
** ------------------------------------------------------------- */
void write_all(int fd, unsigned char *buf)
{
	fd_set writes;
	struct timeval timeout;
	unsigned char *p=buf;
	int i=0, tmp;
	time_t end;

	end=time(NULL)+config_timeout().tv_sec; /*Should we use another value?*/
	
	do{
		FD_ZERO(&writes);
		FD_SET(fd, &writes);
		if( (tmp=end-time(NULL))<0) break;
		timeout.tv_sec=tmp;
		timeout.tv_usec=0;

		i=iselect(FD_SETSIZE, NULL, &writes, NULL, &timeout);
		if(i<1) break;
		i=iwrite(fd, p,strlen(p));
		if(i<1) break;
		p+=i;
	} while(i<strlen(p));

	if(i<1){
		perror("writeall");
		write_log(ERROR, "Error writing to socket. Exiting");
		closeandexit(-1);
	}

}

	
