/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  ftp-cmds.c Parsing code for individual commands.
  
  ***************************************/

#include <fcntl.h>
#include "ftp-cmds.h"

void pasv_parse(const unsigned char *cmd, const unsigned char *arg);
void port_parse(const unsigned char *cmd, const unsigned char * arg);
void abor_parse(const unsigned char *cmd, const unsigned char * arg);
void xfer_command(const unsigned char *cmd, const unsigned char *arg);
void pasv_reply(const unsigned char * msg);

static int converting=FALSE; /*Whether we are doing active<-->passive conversion.*/

void ftpcmds_init()
{
	static cmd_struct list[]  = {  /*Pinched in part SUSE proxy suite! */
		{ "PORT", port_parse },
		{ "PASV", pasv_parse },
		{ "ABOR", abor_parse },
		{ "USER", send_command },
		{ "PASS", send_command },
		{ "ACCT", send_command },
		{ "CWD",  send_command },
		{ "CDUP", send_command },
		{ "SMNT", send_command },
		{ "QUIT", send_command },
		{ "REIN", send_command },
		{ "TYPE", send_command },
		{ "STRU", send_command },
		{ "MODE", send_command },
		{ "RETR", xfer_command },
		{ "STOR", xfer_command },
		{ "STOU", xfer_command },
		{ "APPE", xfer_command },
		{ "ALLO", send_command },
		{ "REST", send_command },
		{ "RNFR", send_command },
		{ "RNTO", send_command },
		{ "DELE", send_command },
		{ "RMD",  send_command },
		{ "MKD",  send_command },
		{ "PWD",  send_command },
		{ "LIST", xfer_command },
		{ "NLST", xfer_command },
		{ "SITE", send_command },
		{ "SYST", send_command },
		{ "STAT", send_command },
		{ "HELP", send_command },
		{ "NOOP", send_command },
		{ "SIZE", send_command },	/* Not found in RFC 959	*/
		{ "MDTM", send_command },
		{ "MLFL", send_command },
		{ "MAIL", send_command },
		{ "MSND", send_command },
		{ "MSOM", send_command },
		{ "MSAM", send_command },
		{ "MRSQ", send_command },
		{ "MRCP", send_command },
		{ "XCWD", send_command },
		{ "XMKD", send_command },
		{ "XRMD", send_command },
		{ "XPWD", send_command },
		{ "XCUP", send_command },
#if 0
		{ "APSV", send_command },	/* As per RFC 1579	*/
#endif
		{ "", 0}	
	};

	ftp_cmds=list;
}

/* ------------------------------------------------------------- **
** Parse the PORT command in arg and store the client's data listening
** port. Either send out a PASV instead, or open a port of our own
** and send this to the server in a rewritten PORT command.
** ------------------------------------------------------------- */
void port_parse(const unsigned char * cmd, const unsigned char *arg)
{
	info->client_data.address=extract_address(arg);
	if(info->client_data.address.sin_addr.s_addr==0){
		send_message(500,"Bad PORT command");
		return;
	}

	if(config_usesameaddress()){
		if(info->client_data.address.sin_addr.s_addr 
		   != info->client_control.address.sin_addr.s_addr){
			send_message(500,"Bad PORT command");
			return;
		}
	}
		
	if(config_bouncedefend()){
		if(ntohs(info->client_data.address.sin_port)<1024){
			send_message(500,"Bad PORT command");
			return;
		}
	}

	if(config_apconv()){
		debug("Rewritten PORT command to PASV\n");
		send_command("PASV","");
		converting=TRUE;
	} else {
		unsigned char newbuf[40];
		int a1,a2,a3,a4,p1,p2;
		struct sockaddr_in listenaddr;
		socklen_t len;

		info->state |= ACTIVE;
		close(info->listen);
		info->state &= ~LISTEN;
		
		len=sizeof(listenaddr);
		getsockname(info->server_control.fd, &listenaddr, &len);
		listenaddr.sin_port=0;
		
		info->listen=listen_on_socket(&listenaddr, ACTV);
		
		if(info->listen == -1){
			send_message(451, "Proxy unable to comply.");
			return;
		} else info->state |= LISTEN;

		n2com(listenaddr,&a1,&a2,&a3,&a4,&p1,&p2);
		
		snprintf(newbuf,40,"%d,%d,%d,%d,%d,%d",a1,a2,a3,a4,p1,p2);
		
		debug("  Rewritten PORT command:\n");
		
		send_command(cmd, newbuf);
	}
}

/* ------------------------------------------------------------- **
** Intercepted a PASV command. 
** ------------------------------------------------------------- */
void pasv_parse(const unsigned char *cmd, const unsigned char *arg)
{
	debug("  Intercepted a PASV command\n");

	info->state &= ~ACTIVE;
	close(info->listen);
	info->state &= ~LISTEN;
	converting=FALSE;

	send_command(cmd, arg );
}

/* ------------------------------------------------------------- **
** Intercepted an ABOR -- we need to send telnet IPs etc. 
** ------------------------------------------------------------- */
void abor_parse(const unsigned char *cmd, const unsigned char *arg)
{
	unsigned char tmp[3];
	tmp[0]=IAC;
	tmp[1]=IP;
	tmp[2]=IAC;
	send(info->server_control.fd,tmp,3,MSG_OOB);
	tmp[0]=DM;
	send(info->server_control.fd,tmp,1,0);
	send_command(cmd, arg);
}

/* ------------------------------------------------------------- **
** Commands that require a data stream.
** ------------------------------------------------------------- */
void xfer_command(const unsigned char *cmd, const unsigned char *arg)
{
	if(converting){
		debug2("Connecting to both data streams for %s command\n",cmd);
		info->client_data.fd=connect_to_socket(info->client_data.address, ACTV);
		if(info->client_data.fd == -1) quit(NO_CLIENT_DATA);

		info->server_data.fd=connect_to_socket(info->server_data.address, PASV);
		if(info->server_data.fd == -1) quit(NO_SERVER_DATA);

		info->server_data.buflen=info->client_data.buflen=0;
		info->state |= CONNECTED;
		
		fcntl(info->client_data.fd, F_SETFL, O_NONBLOCK);
		fcntl(info->server_data.fd, F_SETFL, O_NONBLOCK);
	}	

	send_command(cmd, arg);
}

/* ------------------------------------------------------------- **
** Parse the 227 reply message. Either: a) We are doing
** active->passive conversion - store the port it gives and return OK
** to client's PORT command. b) We are transparently proxying the data
** connection - send the 227 through unchanged, and do a
** intercept_listen() for when the client tries to connect. c) We are
** doing neither - listen on a port of our own and rewrite the 227
** with that.
** ------------------------------------------------------------- */
void pasv_reply(const unsigned char * msg)
{
	unsigned char newbuf[60];
	int a1,a2,a3,a4,p1,p2;
	struct sockaddr_in tmp;

	info->server_data.address=extract_address(msg);
	if(info->server_data.address.sin_addr.s_addr==0) {
		send_message(500, "Bad passive command from server");
		return;
	}

	if(config_usesameaddress()){
		if(info->server_data.address.sin_addr.s_addr != 
		   info->server_control.address.sin_addr.s_addr){
			send_message(500,"Bad PASV command");
			return;
		}
	}

	if(converting){ /*Merely store and wait.*/
		debug("Rewriting 227 reply.\n");
		send_message(200,"PORT command OK.");
		return;
	}

#if 0
	if(config_transdata()){
		get_local_address(&tmp);
		info->listen=intercept_listen(info->server_data.address, tmp,  PASV);
		if(info->listen!=-1) {
			send_message(227, msg);
			return;
		}
		debug("Intercept_listen failed. Rewriting PORT command\n");
	}
#endif

	get_local_address(&tmp);
	tmp.sin_port=0;
	
	info->listen=listen_on_socket(&tmp, PASV);

	if(info->listen == -1) {
                send_message(451, "Screwed up pasv command.");
                return;
        } else info->state |= LISTEN;		

        n2com(tmp,&a1,&a2,&a3,&a4,&p1,&p2);

        snprintf(newbuf,60,"Entering Passive Mode (%d,%d,%d,%d,%d,%d)",
		 a1,a2,a3,a4,p1,p2);

        debug("  Rewritten 227 reply:\n");

        send_message(227, newbuf);
	info->state &= ~ACTIVE; /* FIXME: Do we want this here? */
}
