/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

ntp.c -- non transparent proxying stuff

Overview: 

ntp_changedest gets called just before the proxy connects to the
remote server. It sends a welcome to the client and reads a reply of
the form "USER username[@host[:port]]". If necessary we then change
the destination (in the global "info" structure), save the hostname,
and return.

We also need get called(ntp_senduser) when the remote server sends us
a 220 welcome message. If we have already welcomed the client we send
the login name we got then to the server and return TRUE. Otherwise we
return FALSE and the main proxy code will forward the welcome on to
the client.

***************************************/

#include <sys/ioctl.h>
#include <netdb.h>
#include "common.h"
#include "control.h"

extern int extract_clientcmd(unsigned char * buf, unsigned char * cmd, unsigned char * arg);
void parse_client_cmd(unsigned char *cmd, unsigned char * arg);

void parseuser(void);
void read_username(void);

static char user[MAX_LINE_LEN];
static int working=FALSE;

/* ------------------------------------------------------------- **
**  Called before the proxy connects to the server.
**  
**  Send client "220 Send Login", read and parse reply.
**  ------------------------------------------------------------- */
void ntp_changedest(void)
{
	/*FIXME maybe we should pass config_ntp() some more info to
	 * help it decide. Should we only allow ntp if the user has
	 * connected directly to the proxy rather than being sent here
	 * transparently? */
	if(!config_ntp()) return;

	working=TRUE;

	send_message(220, "Frox transparent ftp proxy. Login with username[@host[:port]]");
	read_username();
	parseuser();
}

/* ------------------------------------------------------------- **
**  If we have a username send it to the server.
**  ------------------------------------------------------------- */
int ntp_senduser(int code)
{
	if(!working) return(FALSE);
	if(working && code!=220) return(TRUE); /*Discard Multiline*/

	working=FALSE;
	parse_client_cmd("USER", user); /*Make sure we go back
					     * through cache code if
					     * necessary*/
	return(TRUE);
}

/* ------------------------------------------------------------- **
**  Parse the user command, resolve the hostname if present, and do
**  security checks. If all ok alter info->server_control.address
**  ------------------------------------------------------------- */
void parseuser()
{
	char *host=NULL, *port=NULL;
	struct hostent* hostinfo;

	debug2("Parsing %s\n", user);
	
	if( (host=strchr(user, '@')) == NULL ) return;
	*host++=0;
	if( (port=strchr(host, ':')) != NULL ) *port++=0;

	debug2("  Host=%s",host);
	debug2(", port=%s\n", port);

	hostinfo=gethostbyname(host);
	if(!hostinfo){
		send_message(501,"Can't find that host");
		quit(SERVER_CLOSE); /*Not really true! */
	}
	if(hostinfo -> h_addrtype != AF_INET){
		send_message(501,"Invalid host");
		quit(SERVER_CLOSE); /*Not really true! */
	}

	info->server_control.address.sin_addr=*((struct in_addr *) hostinfo->h_addr_list[0]);
	if(port!=NULL) info->server_control.address.sin_port=htons(atoi(port));

	/* FIXME security checks. Probably list of allowed ports from
	 * config file. */
}

/* ------------------------------------------------------------- **
**  Mainly lifted from client_control_forward(). Wait for the client
**  to reply and store the username. If it doesn't reply with USER
**  then ask again.
**  ------------------------------------------------------------- */
void read_username()
{
	int i, j;
	unsigned char cmdbuf[5];
	do{
		fd_set reads;
		struct timeval timeout;
		
		FD_ZERO(&reads);
		FD_SET(info->client_control.fd,&reads);
		timeout=config_timeout();
		if(iselect(FD_SETSIZE, &reads, NULL, NULL, &timeout)==0)
			quit(TIMEOUT);
		
		do{
			ioctl(info->client_control.fd, FIONREAD, &i);
		}while(i<0 && errno==EINTR);

		if(i<1) quit(CLIENT_CLOSE);

		j=strlen(info->client_control.buf);
		i = iread(info->client_control.fd, info->client_control.buf+j,
			  (i > (BUF_LEN-1) - j ? (BUF_LEN-1) - j : i) );
		info->client_control.buf[i+j]=0;

		i=extract_clientcmd(info->client_control.buf, cmdbuf, user);
		if(i<0) {
			send_message(500,"You're sending me rubbish. Goodbye!");
			quit(CLIENT_RUBBISH);
		}
		if(strncasecmp(cmdbuf,"USER", 4))
			send_message(530, "Please login with USER first");

	}while(i==1 || strncasecmp(cmdbuf,"USER", 4) );
}
